#include <iostream>
#include <fstream>
#include <sstream>
#include <set>
#include <queue>
#include <map>
#include <vector>
#include <string>
using namespace std;

#include "../Common.h"
#include "../Simulator/Simulator.h"
#include "../Simulator/Node.h"

enum InsnType {
	SEND,
	COMP,
	REQU,
	WAIT
};

struct DFGnode {
	int p;               // belongs to which partition
	int level;
	InsnType t;
	DFGnode* parent1;
	DFGnode* parent2;
	set<DFGnode*> children;
	string name;
};


struct instruction {
	InsnType operand;
	int  level;
	DFGnode* output;
	DFGnode* input1;
	DFGnode* input2;
};

//string = the token string parsed from trafic file
//alternatively, instead of string, we can use the Node::getNodeID(),
//which already provides unique numbers to each variable in the traffic file
map<string, DFGnode*> nodeMap;

void addtoDFGMap(string insn);

void readTrafficFile(char* fileName) {
	ifstream inputFile(fileName);

	// Make sure the file is opened
	if (!inputFile.is_open()) {
		printf("ERROR: Can't open file %s\n", fileName);
		exit(-1);
	}

	string line;
	getline(inputFile, line);
	Debug << line << endl;

	while (inputFile.good()) {
		if (line.compare("") == 0) {
			// Do nothing, it is an empty line
		}
		else if (line.substr(0,2).compare("//") == 0) {
			// Do nothing, it is a comment
		}
		else if (line.substr(0,5).compare("#node") == 0) {
			// It is a directive indicating the following instructions are for which node
			// Do nothing
		}
		else {
			// it is an instruction
			addtoDFGMap(line);
		}

		getline(inputFile, line);
		Debug << line << endl;
	}
}

void addtoDFGMap(string insn) {
	// Read in the tokens
	stringstream ss(insn);
	string tokens[4];
	for (int i=0; i < 4 && ss >> tokens[i]; i++);

	if (tokens[0].compare("send") == 0) {

		assert(nodeMap[tokens[2]] != NULL);  // Send out data must mean you have the data already

		// Find a dummy name that is free
		char c;
		for (c = '!'; nodeMap[tokens[2] + "RO" + c] != NULL; c++);
		string name = tokens[2] + "RO" + c;
		nodeMap[name] = new DFGnode();

		nodeMap[name]->name    = tokens[2];
		nodeMap[name]->parent1 = nodeMap[tokens[2]];
		nodeMap[name]->parent2 = NULL;
		nodeMap[name]->t  = SEND;

		nodeMap[tokens[2]]->children.insert(nodeMap[name]);
	}
	else if (tokens[0].compare("comp") == 0) {

		// Make sure the parents are there
		if (nodeMap[tokens[2]] == NULL) {
			nodeMap[tokens[2]] = new DFGnode();
			nodeMap[tokens[2]]->children.clear();
			nodeMap[tokens[2]]->level = 0;
			nodeMap[tokens[2]]->name = tokens[2];
			nodeMap[tokens[2]]->p = 0;
			nodeMap[tokens[2]]->parent1 = NULL;
			nodeMap[tokens[2]]->parent2 = NULL;
		}

		if (nodeMap[tokens[3]] == NULL) {
			nodeMap[tokens[3]] = new DFGnode();
			nodeMap[tokens[3]]->children.clear();
			nodeMap[tokens[3]]->level = 0;
			nodeMap[tokens[3]]->name = tokens[3];
			nodeMap[tokens[3]]->p = 0;
			nodeMap[tokens[3]]->parent1 = NULL;
			nodeMap[tokens[3]]->parent2 = NULL;
		}

		// Create a new ID
		if (nodeMap[tokens[1]] == NULL) {
			nodeMap[tokens[1]] = new DFGnode();
			nodeMap[tokens[1]]->name = tokens[1];
			nodeMap[tokens[1]]->parent1 = nodeMap[tokens[2]];
			nodeMap[tokens[1]]->parent2 = nodeMap[tokens[3]];
			nodeMap[tokens[1]]->t  = COMP;

			//append curr node in the 2 parents' children set
			nodeMap[tokens[2]]->children.insert(nodeMap[tokens[1]]);
			nodeMap[tokens[3]]->children.insert(nodeMap[tokens[1]]);
		} else {
			char c;
			for (c = '!'; nodeMap[tokens[1] + "RO" + c] != NULL; c++);
			string name = tokens[1] + "RO" + c;
			nodeMap[name] = new DFGnode();
			nodeMap[name]->name = tokens[1];
			nodeMap[name]->parent1 = nodeMap[tokens[2]];
			nodeMap[name]->parent2 = nodeMap[tokens[3]];
			nodeMap[name]->t  = COMP;

			//append curr node in the 2 parents' children set
			nodeMap[tokens[2]]->children.insert(nodeMap[name]);
			nodeMap[tokens[3]]->children.insert(nodeMap[name]);
		}

	}
	else if (tokens[0].compare("requ") == 0) {
		// do nothing?
	}
	else { //Shouldn't happen
		assert(false);
	}
}

// look for new node from the bigger partition to add to the partition
DFGnode* lookForNewNode (set<DFGnode*>* p) {
	DFGnode* second_choice = *(p->begin());
	int minLevel = _I32_MAX;

	// Go through the Graph for good candidates
	for (set<DFGnode*>::iterator it = (++p->begin()); it != p->end(); it++) {
		// Found a node that is the origin, i.e. the input for the function
		if ((*it)->level < minLevel) {
			for (set<DFGnode*>::iterator itt = (*it)->children.begin(); itt != (*it)->children.end(); itt++) {
				if ((*itt)->p == (*it)->p) {
					second_choice = (*it);
					minLevel = (*itt)->level;
					break;
				}
			}
		}
		if (minLevel == 0)
			break;
		/*	if ((*it)->parent1 == NULL && (*it)->parent2 == NULL)
			return *it;
		else if (second_choice == NULL && ((*it)->parent1 != NULL || (*it)->parent2 != NULL))
			second_choice = (*it); */
	}

	return second_choice;
}

// First guess to partition the graph
// p is the set that needs to be divided in half
// about 1/2 the nodes will go into q
// qnum is the number assigned to the new partition
// the results will have 1/2 the nodes in p, 1/2 the nodes in q
void bipartition(int qnum, set<DFGnode*>* p, set<DFGnode*>* q) {
	queue<DFGnode*> parents;    // Parents of nodes in partition that are not mapped yet
	queue<DFGnode*> readyNodes; // Nodes with both parents mapped
	set<DFGnode*> candidates;   // Nodes with one of the 2 parents mapped
	int l = p->size() / 2;
	int pnum = (*(p->begin()))->p;

	// Add l nodes to the partition
	while(q->size() < p->size()) {
		DFGnode* j;

		// Map the parents first
		if (!parents.empty()) {
			j = parents.front();
			parents.pop();
			// Map nodes with both parents mapped second
		}
		else if (!readyNodes.empty()) {
			j = readyNodes.front();
			readyNodes.pop();
			// Map the children of mapped nodes third
		}
		else if (!candidates.empty()) {
			j = *candidates.begin();
			candidates.erase(candidates.begin());
			// look for a new node to map if all else failed
		}
		else  {
			j = lookForNewNode(p);
		}

		j->p = qnum;
		Debug << j->name << endl;
		p->erase(j);
		q->insert(j);

		// Add the parents if they are not in a partition yet
		if (j->parent1 != NULL && j->parent1->p == pnum)
			parents.push(j->parent1);

		if (j->parent2 != NULL && j->parent2->p == pnum)
			parents.push(j->parent2);

		// Add its children to candidates
		for (set<DFGnode*>::iterator it = j->children.begin(); it != j->children.end(); it++) {
			if ((*it)->p == pnum) {
				pair<set<DFGnode*>::iterator,bool> ret = candidates.insert(*it);

				// if the candidate is already mapped or it is a send instruction, then upgrade it to readyNode
				if(ret.second == false || (*it)->t == SEND) {
					readyNodes.push(*it);
					candidates.erase(ret.first);
				}
			}
		}
	}
}

// Calculate the level of DFGnode n
void calculateLevel(DFGnode* n) {
	if (n->level > 0)
		return;

	if (n->parent1 == NULL && n->parent2 == NULL) {
		n->level = 0;
		return;
	}

	int l = 0;

	if (n->parent1 != NULL) {
		calculateLevel(n->parent1);
		l = n->parent1->level;
	}

	if (n->parent2 != NULL) {
		calculateLevel(n->parent2);
		l = max(l,n->parent2->level);
	}

	n->level = l + 1;
}

void bipartition_main(int start, int end, set<DFGnode*>* partition) {

	// No more further partition
	if (end <= start)
		return;

	// New Partition number
	int qnum = start + (end - start + 1) / 2;
	printf("P%d size %d\n", start, partition[start].size());
	bipartition(qnum,&partition[start],&partition[qnum]);
	printf("P%d size %d P%d size %d\n", start, partition[start].size(), qnum, partition[qnum].size());

	bipartition_main(start, qnum - 1, partition);
	bipartition_main(qnum, end, partition);
	
}

/**
* Arguments Expected:
* argv[0] - app name
* argv[1] - Number of Partitions
* argv[2] - Data flow graph in text format
* argv[3] - Configuration File
* argv[4] - Output File
*/
int main(int argc, char **argv) {
	int numPartition = atoi(argv[1]);
	int maxLevel = 0;

	// Read in the Data Flow Graph
	assert(argc == 5);
	set<DFGnode*>* partition = new set<DFGnode*>[numPartition];
	queue<DFGnode*> qlevel;
	deque<instruction>* insn = new deque<instruction>[numPartition + 1];

	readTrafficFile(argv[2]);

	// Insert all the nodes into one partition
	for (map<string,DFGnode*>::iterator it = nodeMap.begin(); it != nodeMap.end(); it++) {
		DFGnode* tmp = ((*it).second);
		tmp->p = 0;
		tmp->level = 0;
		partition[0].insert(tmp);

		// Remember data sources
		if (tmp->parent1 == 0 && tmp->parent2 ==0)
			qlevel.push(tmp);
	}

	// Calculate the level of every node
	while (!qlevel.empty()) {
		DFGnode* n = qlevel.front();
		qlevel.pop();

		// Only calculate if the level haven't been calculated
		calculateLevel(n);
		maxLevel = max(maxLevel,n->level);

		//	Debug << n->name << " Level " << n->level << endl;

		// Add the children to the queue, only if its level is unknown
		for (set<DFGnode*>::iterator it = n->children.begin(); it != n->children.end(); it++){
			if ((*it)->level == 0) {
				qlevel.push(*it);
			}
		}
	}

	// Partition the graph

	bipartition_main(0,numPartition-1,partition);
	//bipartition(2,&partition[0],&partition[2]);
	//bipartition(1,&partition[0],&partition[1]);
	//bipartition(3,&partition[2],&partition[3]);

	// Print out the partition
#ifdef DEBUG
	for (int i = 0; i < numPartition; i++) {
		Debug << "Parition " << i << endl;
		for (set<DFGnode*>::iterator it = partition[i].begin(); it != partition[i].end(); it++) {
			//if ((*it)->children.empty()) {
			Debug << "Data " << (*it)->name << " Parents: " << endl;
			if ((*it)->parent1 != NULL)
				Debug << (*it)->parent1->name << " partition " << (*it)->parent1->p << endl;
			if ((*it)->parent2 != NULL)
				Debug << (*it)->parent2->name << " partition " << (*it)->parent2->p << endl;
			Debug << endl;  
			//	}
		}
	}
#endif

	// Create the topology
	//	Simulator Simulation(argv[3]);

	// Add send commands to main memory controller
	// Bucket sort the DFGnode* by their level
	vector<DFGnode*>*  nodeByLevel = new vector<DFGnode*>[maxLevel+1];
	for (map<string,DFGnode*>::iterator it = nodeMap.begin(); it != nodeMap.end(); it++) {
		DFGnode* tmp = ((*it).second);
		nodeByLevel[tmp->level].push_back(tmp);
	}

	// Send out data sources to its parent
	for (vector<DFGnode*>::iterator it = nodeByLevel[0].begin(); it != nodeByLevel[0].end(); it++) {
		DFGnode* tmp = ((*it));
		instruction in;
		in.level = 0;
		in.operand = SEND;
		in.input1 = tmp; 
		in.input2 = 0;
		in.output = (DFGnode*) (tmp->p + 1);
		insn[0].push_back(in);

		// Tell its children to request for the data if they are not in the same partition
		set<DFGnode*> sendTo;
		in.level = 0;
		in.operand = REQU;
		in.output = (*it);
		in.input1 = (*it); 
		in.input2 = 0;
		for (set<DFGnode*>::iterator itt = (*it)->children.begin(); itt != (*it)->children.end(); itt++) {
			if ((*itt)->p != (*it)->p)
				sendTo.insert((*itt));
		}

		for (set<DFGnode*>::iterator itt = sendTo.begin(); itt != sendTo.end(); itt++) {
			in.input2 = (*itt);
			insn[(*itt)->p+1].push_back(in);
		}

	}

	// Initialize the instructions for the partitions
	for (int i = 1; i <= maxLevel; i++) {
		// map the lower level instructions first
		Debug << "Generating instructions for Level " << i << endl;
		for (vector<DFGnode*>::iterator it = nodeByLevel[i].begin(); it != nodeByLevel[i].end(); it++) {
			int level = i;
			int part = (*it)->p;

			instruction in;

			switch ((*it)->t)
			{
			case SEND:
				// Send to main memory
				in.level = level;
				in.operand = SEND;
				in.output = (DFGnode*) 0;
				in.input1 = (*it); 
				in.input2 = 0;
				insn[(*it)->p+1].push_back(in);

				// wait for the data in main memory
				in.level = level;
				in.operand = WAIT;
				in.output = (DFGnode*) 0;
				in.input1 = (*it); 
				in.input2 = 0;
				insn[0].push_back(in);
				break;
			case COMP:
				in.input1  = (*it)->parent1;
				in.input2  = (*it)->parent2;
				in.level   = level;
				in.operand = COMP;
				in.output  = (*it);
				insn[part+1].push_back(in);
				break;
			default:
				assert(false);
				break;
			}

			// Tell the other partition to ask for the data if needed
			set<DFGnode*> sendTo;
			in.level = level;
			in.operand = REQU;
			in.output = (*it);
			in.input1 = (*it); 
			in.input2 = 0;
			for (set<DFGnode*>::iterator itt = (*it)->children.begin(); itt != (*it)->children.end(); itt++) {
				if ((*itt)->p != part)
					sendTo.insert((*itt));
			}

			for (set<DFGnode*>::iterator itt = sendTo.begin(); itt != sendTo.end(); itt++) {
				in.input2 = (*itt);
				insn[(*itt)->p+1].push_back(in);
			}

		}		
	}

	// Print out generated instructions
	ofstream outputFile(argv[4]);
	if (!outputFile.is_open()) {
		printf("Unable to open output file for writing.\n");
		exit(1);
	}

	for (int i = 0; i < numPartition+1; i++) {
		outputFile << "#node " << i << endl;
		for (deque<instruction>::iterator it = insn[i].begin(); it != insn[i].end(); it++) {
			switch ((*it).operand)
			{
			case WAIT:
				Debug      << "wait " << (*it).input1->name << endl;
				outputFile << "wait " << (*it).input1->name << endl;
				break;
			case SEND:
				Debug      << "send " << (int)(*it).output << " " << (*it).input1->name << endl;
				outputFile << "send " << (int)(*it).output << " " << (*it).input1->name << endl;
				break;
			case COMP:
				Debug      << "comp " << (*it).output->name << " " << (*it).input1->name << " " << (*it).input2->name << endl;
				outputFile << "comp " << (*it).output->name << " " << (*it).input1->name << " " << (*it).input2->name << endl;
				break;
			case REQU:
				Debug      << "requ " << (*it).output->p+1 << " " << (*it).input1->name << " " << (*it).input2->p+1 << endl;
				outputFile << "requ " << (*it).output->p+1 << " " << (*it).input1->name << " " << (*it).input2->p+1 << endl;
				break;
			default:
				assert(false); // shouldn't happen
				break;
			}
		}
	}

	outputFile << "// Completed with " << numPartition << endl;

	// Clean up
	for (map<string,DFGnode*>::iterator it = nodeMap.begin(); it != nodeMap.end(); it++) {
		if ((*it).second != NULL)
			delete (*it).second;
	}
	nodeMap.clear();

	for (int i = 0; i < numPartition; i++)
		partition[i].clear();
	delete [] partition;

	return 0;
}
