// block_lu.cpp : main project file.

/*
How to Use:
Run from command line: <name of app> <input matrix file> <output lu matrix file>

Format of <input matrix file> :
<number of rows/cols in matrix> <block size>
<row 1 of matrix separated by spaces>
...
<last row of matrix separated by spaces>

For the purpose of this program, consider the following conventions:
matrixSize	 = number of rows (= number of columns) in the whole matrix
blkSize	     = number of rows (= number of columns) in a block (sub-matrix)
numOfDimblks = number of blocks in a row = number of blocks in a column = N / blkSize

Example:

matrix						block numbers
-------------------			-------------------
| # # | # # | # # |			|     |     |     |
| # # | # # | # # |			|  0  |  3  |  6  |
-------------------			-------------------
| # # | # # | # # |			|     |     |     |
| # # | # # | # # |			|  1  |  4  |  7  |
-------------------			-------------------	
| # # | # # | # # |			|     |     |     |
| # # | # # | # # |			|  2  |  5  |  8  |
-------------------			-------------------

matrixSize=6, blkSize=2, numOfDimblks=3
*/

#include "helper.h"
#include <fstream>
#include <map>

float* root;
map<float *, int> m;

/**
* Returns the block type.
*    1 - upper-leftmost block (depends on none)
*    2 - leftmost column blocks discluding block 0 (depends on block 0)
*    3 - topmost row blocks discluding block 0 (depends on the leftmost block on the same row)
*    4 - the rest (depends on the topmost block on the same column AND leftmost block on the same row)
*
* @param blkNum See diagram above. The blocks are numbered vertically, starting from the upper left and moving right.
* @param numOfDimBlks Number of blocks in a row = number of blocks in a column.
*/
uint blockType(const uint &blkNum, const uint numOfDimBlks) {	
	if (blkNum == 0)                              return (1); // case 1
	else if (blkNum>=1 && blkNum<=numOfDimBlks-1) return (2); // case 2
	else if (blkNum % numOfDimBlks == 0)          return (3); // case 3
	else                                          return (4); // case 4
}

// Prints out the matrixSize x matrixSize matrix onto the console
void printMatrix(float *matrix, uint matrixSize) {
	for (uint i=0; i<matrixSize; i++) {
		for (uint j=0; j<matrixSize; j++)
			printf("%f ", matrix[i*matrixSize+j]);		
		printf("\n");
	}
}

float *ReadMatrixFromFile(const char* inputFileName, uint &matrixSize, uint &blkSize) {
	ifstream inputFile(inputFileName);
	if (!inputFile.is_open()) {
		printf("Unable to open input matrix file for reading.\n");
		exit(1);
	}

	inputFile >> matrixSize >> blkSize;

	if (matrixSize % blkSize != 0) {
		printf("Matrix size must be divisible by block size.\n");
		exit(1);
	}

	float *matrix = new float[matrixSize * matrixSize];

	for (uint i=0; i<matrixSize*matrixSize; i++)
		inputFile >> matrix[i];

	inputFile.close();
	return matrix;
}

void WriteMatrixToFile(const char* outputFileName, float *matrix, const uint matrixSize) {
	ofstream outputFile(outputFileName);
	if (!outputFile.is_open()) {
		printf("Unable to open output matrix file for writing.\n");
		exit(1);
	}

	outputFile << matrixSize << "\n";

	for (uint i=0; i<matrixSize; i++) {
		for (uint j=0; j<matrixSize; j++)
			outputFile << matrix[i*matrixSize+j] << " ";
		outputFile << "\n";
	}

	outputFile.close();
}

uint main(uint argc, char **argv) {
	if (argc != 3) {
		printf("Invalid arguments provided.\n");
		printf("Required arguments: <path to input matrix file> <path to output matrix file>");
		exit(1);
	}

	//matrixSize = number of rows = number of columns in the matrix
	//blkSize = the size of each block
	uint matrixSize, blkSize;

	float *matrix = ReadMatrixFromFile(argv[1], matrixSize, blkSize);

	ofstream outputFile(argv[2]);
	if (!outputFile.is_open()) {
		printf("Unable to open output matrix file for writing.\n");
		exit(1);
	}

	outputFile << "//LU parameters Used:"
		<< "\n//matrix size = " << matrixSize
		<< "\n//block size = "  << blkSize << "\n\n";

	float *submatrix = matrix;
	root = matrix;

	//numOfDimBlks = number of blocks in a row/column
	uint numOfDimBlks = matrixSize / blkSize, numOfIterations = numOfDimBlks;

	for (uint iteration=1; iteration<=numOfIterations; iteration++) {
		//printf("Iteration: %d\n", iteration);

		float *block = submatrix, *top_block = submatrix, *left_block = submatrix;

		for (uint blkNum=0; blkNum<numOfDimBlks*numOfDimBlks; blkNum++) {
			uint blkType = blockType(blkNum, numOfDimBlks);
			//printf("Processing block %d  which is of type %d\n", blkNum, blkType);

			if (blkType == 1) {
				LUDecompType1(block, blkSize, matrixSize, outputFile);
				// Send back the results to main memory	
				// Note that 1st row doesn't change for BlockType 1
				for (uint i = 1; i < blkSize; i++) {
					for (uint j = 0; j < blkSize; j++) {
						outputFile << "send 0 "
							<< matrixIndex(&block[i*matrixSize + j],matrixSize).c_str()
							<< m[&block[i*matrixSize + j]] << endl;
					}
				}
			}
			else if (blkType == 2) {
				block += matrixSize*blkSize;   //move 1 block down the column
				LUDecompType2(block, top_block, blkSize, matrixSize, outputFile);
				// Send back the results to main memory			
				for (uint i = 0; i < blkSize; i++) {
					for (uint j = 0; j < blkSize; j++) {
						outputFile << "send 0 "
							<< matrixIndex(&block[i*matrixSize + j],matrixSize).c_str()
							<< m[&block[i*matrixSize + j]] << endl;
					}
				}
			}
			else if (blkType == 3) {
				block -= (numOfDimBlks-1)*matrixSize*blkSize; //to get back to the top of the column of the submatrix
				block += blkSize;                             //to move to the column to the right
				left_block = submatrix;                       //left_block is reset
				LUDecompType3(block, left_block, blkSize, matrixSize, outputFile);

				// Send back the results to main memory
				// Note that the 1st row doesn't change in BlockType 3
				for (uint i = 1; i < blkSize; i++) {
					for (uint j = 0; j < blkSize; j++) {
						outputFile << "send 0 "
							<< matrixIndex(&block[i*matrixSize + j],matrixSize).c_str()
							<< m[&block[i*matrixSize + j]] << endl;
					}
				}

				top_block += blkSize;          //ready for the rest of the blocks in that column (which are block type 4)
			}
			else if (blkType == 4) {
				block += matrixSize*blkSize;       //move 1 block down the column
				left_block += matrixSize*blkSize;  //move 1 block down the column
				LUDecompType4(block, top_block, left_block, blkSize, matrixSize, outputFile);

				// Send back the results to main memory			
				for (uint i = 0; i < blkSize; i++) {
					for (uint j = 0; j < blkSize; j++) {
						outputFile << "send 0 "
							<< matrixIndex(&block[i*matrixSize + j],matrixSize).c_str()
							<< m[&block[i*matrixSize + j]] << endl;
					}
				}
			}
		}

		submatrix += (matrixSize*blkSize + blkSize);
		numOfDimBlks--;
	}

	//WriteMatrixToFile(argv[2], matrix, matrixSize);

	outputFile.close();
	delete[] matrix;
	return 0;
}
