#include <iostream>
#include <sstream>
using namespace std;

#include "MemoryCntrl.h"

// Constructor
MemoryCntrl::MemoryCntrl() : Node() {
	NumberPacketSendPerCycle = 16;
	NumberPacketReceivePerCycle = 16;
	dataSent = 0;
	dataReceived = 0;
}

// Destructor
MemoryCntrl::~MemoryCntrl() {
	waitForData.clear();	
	instructions.clear();
	availableData.clear();
	pendingPackets.clear();
}

void MemoryCntrl::setupNode(Config* config, uint id) {
	Node::setupNode(config, id);

	NumberPacketSendPerCycle     = config->getUInt("MemoryCntrl->NumberPacketSendPerCycle");
	NumberPacketReceivePerCycle  = config->getUInt("MemoryCntrl->NumberPacketReceivePerCycle");
	
	*outputFile << "DataSent" << ", " << "DataReceive" << endl;
}

bool MemoryCntrl::receive(Packet p) {
	if (NumberPacketReceivePerCycle > incomingPackets->size()) {
		incomingPackets->push(p);
		return (true);
	}
	return (false);
}

bool MemoryCntrl::send(set<Node*>& usedLinks, Packet p) {
	if (usedLinks.find(routingTable[p.destNodeID]) == usedLinks.end()
		&& routingTable[p.destNodeID]->receive(p)) {
			// Remove the link if the destination can receive the packet
			usedLinks.insert(routingTable[p.destNodeID]);

			#ifdef DEBUG
				Debug << "send " << p.destNodeID << " " << IDtoName[p.dataID] << endl;
			#endif
			return true;
	}

	return false;
}

// Setup before starting the step again
void MemoryCntrl::preStep() {
	// Swap the arrivedPacket queue and incomingPacket queue
	// This way, incomingPacket is empty once more and
	// arrivedPackets are the incomingPacket of previous step
	queue<Packet>* tmp = arrivedPackets;
	arrivedPackets = incomingPackets;
	incomingPackets = tmp;
}

// Tell the node to do one cycle of work
void MemoryCntrl::step() {
	Debug << "MemCntrl->step()\n";
	uint sendToken = NumberPacketSendPerCycle;
	set<Node*> usedLinks;

	// Look for Wait instructions
	deque<MemoryCntrlInsn>::iterator insnIt = instructions.begin();
	while (insnIt != instructions.end()) {
		// Try executing the instruction
		switch (insnIt->t) {
			case MCIT_WAIT:
				waitForData.insert(insnIt->input1);
				insnIt = instructions.erase(insnIt);
				break;

			default:
				insnIt++;
				break;
		}
	}

	// Process incoming Packets, relay them if possible, store them if they have to stay 
	while (!arrivedPackets->empty()) {
		Packet p = arrivedPackets->front(); //front returns the element that pop will remove
		arrivedPackets->pop(); //pop has no return type

		// Consume the packet if it is for this node
		if (p.destNodeID == nodeId) {
			switch (p.t)
			{
				case PKT_SEND:  // SEND Process
					if (waitForData.find(p.dataID) != waitForData.end()) {
						availableData.insert(p.dataID);
						waitForData.erase(p.dataID);
						dataReceived++;
						#ifdef DEBUG
							Debug << "received " << IDtoName[p.dataID] << endl;
						#endif
					}
					else {
						#ifdef DEBUG
							Debug << "received " << IDtoName[p.dataID] << endl;
						#endif
						assert(false);  // Node is not waiting for this piece of data
					}
					break;

				case PKT_REQUEST: // REQUEST Process 
					p.t = PKT_SEND;
					p.destNodeID = p.returnNodeID;
					p.returnNodeID = 0;
					pendingPackets.push_back(p);
					break;

				default:
					break;
			}
		}
		else {
			// Relay the packet if the packet is not for this node
			if (!send(usedLinks,p)) {
				// Push the packet to outgoing packet queue
				outgoingPackets.push_back(p);
			}
		}
	}

	// Process Outgoing Packets
	vector<Packet>::iterator outPaketIt = outgoingPackets.begin();
	while (outPaketIt != outgoingPackets.end() && sendToken > 0) {
		Packet p = *outPaketIt;
		if (send(usedLinks,p)) {
			outPaketIt = outgoingPackets.erase(outPaketIt);
			dataSent++;
			sendToken--;
		}
		else {
			outPaketIt++;
		}
	}

	// Process Pending Packets
	// Send only the data the controller had waited for
	vector<Packet>::iterator penPaketIt = pendingPackets.begin();
	while (penPaketIt != pendingPackets.end() && sendToken > 0) {
		Packet p = *penPaketIt;
		if (availableData.find(p.dataID) != availableData.end() && send(usedLinks,p)) {
			penPaketIt = pendingPackets.erase(penPaketIt);
			sendToken--;
			dataSent++;
		}
		else {
			penPaketIt++;
		}
	}

	// Process Instruction
	insnIt = instructions.begin();
	while (sendToken > 0 && insnIt != instructions.end()) {
		// Try executing the instruction
		switch (insnIt->t) {
			case MCIT_SEND:
				Packet p;
				p.dataID = insnIt->input1;
				p.destNodeID = insnIt->output1;
				p.t = PKT_SEND;

				// Try sending out
				if (send(usedLinks,p)) {
					sendToken--;
					dataSent++;
					insnIt = instructions.erase(insnIt);
				}
				else {
					insnIt++;
				}
				break;

			default:
				insnIt++;
				break;
		}
	}
}

void MemoryCntrl::addInstruction(string insn) {
	// Read in the tokens
	stringstream ss(insn);
	string tokens[3];
	for (uint i=0; i < 3 && ss >> tokens[i]; i++);

	// Decide what to do depending on the operand 
	MemoryCntrlInsn newInsn;
	if (tokens[0].compare("send") == 0) {
		newInsn.t = MCIT_SEND;
		newInsn.output1 = atoi(tokens[1].c_str());
		newInsn.input1  = Node::getDataID(tokens[2]);
		Debug << "MemCntrl: send " << newInsn.output1 << " " << newInsn.input1 << endl;
	} 
	else if (tokens[0].compare("wait") == 0) {
		newInsn.t = MCIT_WAIT;
		newInsn.input1 = Node::getDataID(tokens[1]);
		Debug << "MemCntrl: wait " << newInsn.input1 << endl;
	}
	else { //Shouldn't happen
		assert(false);
	}

	instructions.push_back(newInsn);
}

//Node is completed when there's no more instructions to do, no more incoming packets,
//and no more outgoing packets to process
bool MemoryCntrl::isCompleted() {
	return (instructions.size() == 0
		&& incomingPackets->size() == 0
		&& outgoingPackets.size() == 0
		&& pendingPackets.size() == 0
		&& waitForData.size() == 0);
}

// Reset counters
void MemoryCntrl::resetCounters() {
	dataReceived = 0;
	dataSent = 0;
}

// print stats
void MemoryCntrl::printStat() {
	*outputFile << dataSent << ", " << dataReceived << endl;
}
