#include <iostream>
#include <sstream>
using namespace std;

#include "SimpleNode.h"

// Constructor
SimpleNode::SimpleNode() : Node() {
	InstructionWindowSize   = 1;
	InstructionsPerCycle    = 1;
	NumberIOPorts           = 4;
	NumberBRAMWritePerCycle = 1;
	NumberBRAMReadPerCycle  = 2;

	resetCounters();
}

// Destructor
SimpleNode::~SimpleNode() {
	availableData.clear();
	ioports.clear();
	tobeWritten.clear();
	instructions.clear();
}

void SimpleNode::resetCounters() {
	requSent = 0;
	dataReceived = 0;
	requReceived = 0;
	dataSent = 0;
	idleCycle = 0;
	compCycle = 0;
	sendCycle = 0 ; 
}

void SimpleNode::printStat () {
	*outputFile << requSent << ", " << requReceived << ", " 
		        << dataSent << ", " << dataReceived << ", "
				<< idleCycle << ", "
				<< compCycle << ", " << sendCycle << endl;
}

void SimpleNode::setupNode(Config* config, uint nodeId) {
	Node::setupNode(config, nodeId);

	InstructionWindowSize   = config->getUInt("SimpleNode->InstructionWindowSize");
	InstructionsPerCycle    = config->getUInt("SimpleNode->InstructionPerCycle");
	NumberIOPorts           = config->getUInt("SimpleNode->NumberIOPorts");
	NumberBRAMWritePerCycle = config->getUInt("SimpleNode->NumberBRAMWritePerCycle");
	NumberBRAMReadPerCycle  = config->getUInt("SimpleNode->NumberBRAMReadPerCycle");
	
	*outputFile << "RequSent, RequReceived, DataSent, DataReceived, IDLE, COMP, SEND" << endl; 
}

bool SimpleNode::receive(Packet p) {
	if (NumberIOPorts > ioports.size() + incomingPackets->size()){
		incomingPackets->push(p);
		return (true);
	}
	return (false);
}

bool SimpleNode::send(set<Node*>& usedLinks, Packet p) {
	if (usedLinks.find(routingTable[p.destNodeID]) == usedLinks.end()
		&& routingTable[p.destNodeID]->receive(p)) {
			// Remove the link if the destination can receive the packet
			usedLinks.insert(routingTable[p.destNodeID]);

			#ifdef DEBUG
				if(p.t == PKT_SEND)
					Debug << "send " << p.destNodeID << " " << IDtoName[p.dataID] << endl;
				else if (p.t == PKT_REQUEST)
					Debug << "requ " << p.destNodeID << " " << IDtoName[p.dataID] << " " << p.returnNodeID << endl;
			#endif
			return true;
	}

	return false;
}

// Setup before starting the step again
void SimpleNode::preStep() {
	// Swap the arrivedPacket queue and incomingPacket queue
	// This way, incomingPacket is empty once more and
	// arrivedPackets are the incomingPacket of previous step
	queue<Packet>* tmp = arrivedPackets;
	arrivedPackets = incomingPackets;
	incomingPackets = tmp;
}

// Tell the node to do one cycle of work
void SimpleNode::step() {
	Debug << "SimpleNode->step()\n";
	uint insnToken = InstructionsPerCycle;
	uint readToken = NumberBRAMReadPerCycle;
	uint writeToken = NumberBRAMWritePerCycle;
	set<Node*> usedLinks;
	bool isSEND = false;  // becomes true when any data is sent
	bool isCOMP = false;  // becomes true when the node consumes 1 instruction or more

	// Process incoming Packets, relay them if possible, store them if they have to stay 
	while (!arrivedPackets->empty()) {
		Packet p = arrivedPackets->front(); //front returns the element that pop will remove
		arrivedPackets->pop(); //pop has no return type

		// Consume the packet if it is for this node
		if (p.destNodeID == nodeId) {
			switch (p.t) {
				case PKT_SEND:  // SEND Process
					if (ioports.size() < NumberIOPorts) {
						ioports.insert(p.dataID);
						tobeWritten.insert(p.dataID);
						dataReceived++;

						#ifdef DEBUG
							Debug << "received " << IDtoName[p.dataID] << endl;
						#endif
					}
					else {
						availableData.insert(p.dataID);
						//assert(writeToken > 0);  // compiler bug?
						//printf("\n(NodeId, WriteTokens): (%d,%d)\n", nodeId,writeToken);
						writeToken--;
						dataReceived++;
					}
					break;

				case PKT_REQUEST: // REQUEST Process 
					p.t = PKT_SEND;
					p.destNodeID = p.returnNodeID;
					p.returnNodeID = 0;
					requReceived++;
					outgoingPackets.push_back(p);
					break;

				default:
					break;
			}

		}
		else {
			// Relay the packet if the packet is not for this node
			switch (p.t) {
				case PKT_REQUEST:  // Request Process, give some effort to see if it already have what they want
					if (ioports.find(p.dataID) != ioports.end() ||
						availableData.find(p.dataID) != availableData.end()) {
							p.t = PKT_SEND;
							p.destNodeID = p.returnNodeID;
							p.returnNodeID = 0;
							requReceived++;
							// Push the packet to outgoing packet queue
							outgoingPackets.push_back(p);
							break;
					}					

				case PKT_SEND:  // SEND Process
					if (!send(usedLinks,p)) {
						// Store the data onto IO port or BRAM
						if (ioports.size() < NumberIOPorts) {
							ioports.insert(p.dataID);
						}
						else {
							availableData.insert(p.dataID);
							writeToken--;
							assert(writeToken >= 0);
						}

						// Push the packet to outgoing packet queue
						outgoingPackets.push_back(p);
					}
					break;

				default:
					assert(false); // shouldn't be here
			}
		}
	}

	// Process Outgoing Packets
	vector<Packet>::iterator outPaketIt = outgoingPackets.begin();
	set<uint> dataRead;
	while (outPaketIt != outgoingPackets.end()) {
		Packet p = *outPaketIt;

		// Just send it out if it is a request
		if (p.t == PKT_REQUEST) {
			if (send(usedLinks,p)) {
				outPaketIt = outgoingPackets.erase(outPaketIt);
				requSent++;
				isSEND = true;
			}
			else {
				outPaketIt++;
			}
		}
		// Try finding the data in the IO ports or read when responding another request
		else if (ioports.find(p.dataID)  != ioports.end() 
			  || dataRead.find(p.dataID) != dataRead.end() ) {
			if (send(usedLinks,p)) {
				outPaketIt = outgoingPackets.erase(outPaketIt);
				dataSent++;
				isSEND = true;
			}
			else {
				outPaketIt++;
			}
		}
		// Try finding the data in BRAM
		else if (readToken > 0 && availableData.find(p.dataID) != availableData.end()) {
			if (send(usedLinks,p)) {
				outPaketIt = outgoingPackets.erase(outPaketIt);
				readToken--;
				dataRead.insert(p.dataID);
				isSEND = true;
				dataSent++;
			}
			else {
				outPaketIt++;
			}
		}
		// The node is still waiting for the outgoing data
		else {
			outPaketIt++;

		}
	}

	// Process Instruction
	deque<SimpleNodeInsn>::iterator insnIt = instructions.begin();
	for (uint i = 0; i < InstructionWindowSize && insnToken > 0 
		&& insnIt != instructions.end() && readToken > 0; i++) {

			bool isInput1Found = false;
			bool isInput2Found = false;
			uint readTokenUsed = 0;

			// Try finding input1 from IO ports first
			if (ioports.find(insnIt->input1) != ioports.end()) {
				isInput1Found = true;
				// Try finding input1 from BRAM
			}
			else if (readToken > readTokenUsed
				&& availableData.find(insnIt->input1) != availableData.end()) {
					isInput1Found = true;
					readTokenUsed++;
			}

			// Try executing the instruction
			switch (insnIt->t) {
				case SNIT_SEND:
					// Check if the input is available on BRAM
					if (isInput1Found) {            
						Packet p;
						p.dataID = insnIt->input1;
						p.destNodeID = insnIt->output1;
						p.t = PKT_SEND;
						p.returnNodeID = 0;

						// Try sending out
						if (send(usedLinks,p)) {
							readToken -= readTokenUsed;
							insnToken--;
							insnIt = instructions.erase(insnIt);
							dataSent++;
							isSEND = true;
						}
						else {
							insnIt++;
						}
					}
					break;

				case SNIT_REQUEST:
					// Don't send out the request if this node already has it
					if (isInput1Found) {
						insnIt = instructions.erase(insnIt);
						if (ioports.find(insnIt->input1) != ioports.end()) {
							tobeWritten.insert(insnIt->input1);
						}
						break;
					}

					Packet p;
					p.dataID = insnIt->input1;
					p.destNodeID = insnIt->output1;
					p.t = PKT_REQUEST;
					p.returnNodeID = nodeId;

					// Try sending out
					if (send(usedLinks,p)) {
						insnToken--;
						insnIt = instructions.erase(insnIt);
						requSent++;
						isSEND = true;
					}
					else {
						insnIt++;
					}
					break;

				case SNIT_COMPUTE:
					// Check if the inputs are available on BRAM
					// and if there are enough resources for a write

					// Try finding input2 from IO ports first
					if (ioports.find(insnIt->input2) != ioports.end()) {
						isInput2Found = true;
						// Try finding input1 from BRAM
					}
					else if (readToken > readTokenUsed
						&& availableData.find(insnIt->input2) != availableData.end()) {
							isInput2Found = true;
							readTokenUsed++;
					}

					// Try writing to IO port or BRAM if both input 1 & 2 are found
					if (isInput1Found && isInput2Found) {
						if (writeToken > 0) {
							availableData.insert(insnIt->output1);
							writeToken--;

							#ifdef DEBUG
								Debug << "comp " << IDtoName[insnIt->output1]
								<<  " "    << IDtoName[insnIt->input1]
								<<  " "    << IDtoName[insnIt->input2]
								<< endl;
							#endif

							insnIt = instructions.erase(insnIt);
							insnToken--;
							isCOMP = true;
							readToken -= readTokenUsed;
						}
						else if (ioports.size() + incomingPackets->size() < NumberIOPorts) {
							ioports.insert(insnIt->output1);
							tobeWritten.insert(insnIt->output1);

							#ifdef DEBUG
								Debug << "comp " << IDtoName[insnIt->output1]
								<<  " "    << IDtoName[insnIt->input1]
								<<  " "    << IDtoName[insnIt->input2]
								<< endl;
							#endif
							
							insnIt = instructions.erase(insnIt);
							insnToken--;
							isCOMP = true;
							readToken -= readTokenUsed;
						}
						else {
							insnIt++;
						}
					}
					else {
						insnIt++;
					}
					break;

				case SNIT_WAIT:
					insnIt++;
					break;

				default:
					insnIt++;
					break;
			}
	}

	// Clear out data from IO ports
	set<uint>::iterator ioPortIt = ioports.begin();
	while (ioPortIt != ioports.end()) {
		uint d = *ioPortIt;
		bool isOutGoing = false;

		// Go through the outgoingPackets list and check if the data is still needed
		for (uint i = 0; i < outgoingPackets.size(); i++) {
			if (d == outgoingPackets[i].dataID) {
				isOutGoing = true;
				break;
			}
		}
		set<uint>::iterator tobeWrittenIt = tobeWritten.begin();
		while (tobeWrittenIt != tobeWritten.end()) {
			if (d == *tobeWrittenIt) {
				if (writeToken > 0) {
					// Write the data to BRAM if there is a free write
					availableData.insert(d);
					tobeWrittenIt = tobeWritten.erase(tobeWrittenIt);
					writeToken--;
					break;
				}
				else {
					isOutGoing = true;
					break;
				}
			} else {
				tobeWrittenIt++;
			}
		}

		// Clear the port if it is not needed any more
		if (!isOutGoing) {
			ioPortIt = ioports.erase(ioPortIt);
		}
		else {
			ioPortIt++;
		}
	}

	// Prevent deadlock by using up a free write to transfer data from IOport to BRAM
	if (!ioports.empty() && writeToken > 0) {
		writeToken--;
		availableData.insert(*ioports.begin());
		ioports.erase(ioports.begin());
	}

	// Determine which type of cycle this is
	if (isCOMP)
		compCycle++;
	else if (isSEND)
		sendCycle++;
	else
		idleCycle++;
}

void SimpleNode::addInstruction(string insn) {
	// Read in the tokens
	stringstream ss(insn);
	string tokens[4];
	for (uint i=0; i < 4 && ss >> tokens[i]; i++);

	// Decide what to do depending on the operand
	SimpleNodeInsn newInsn;
	if (tokens[0].compare("send") == 0) {
		newInsn.t = SNIT_SEND;
		newInsn.output1 = atoi(tokens[1].c_str());
		newInsn.input1  = Node::getDataID(tokens[2]);

		Debug << "SimpleNode: send " << newInsn.output1 << " " << newInsn.input1 << endl;
	} 
	else if (tokens[0].compare("requ") == 0) {
		newInsn.t = SNIT_REQUEST;
		newInsn.output1 = atoi(tokens[1].c_str());
		newInsn.input1  = Node::getDataID(tokens[2]);
		newInsn.input2  = this->nodeId;

		Debug << "SimpleNode: requ " << newInsn.output1 << " " << newInsn.input1 << " " << newInsn.input2 << endl;
	} 
	else if (tokens[0].compare("comp") == 0) {
		newInsn.t = SNIT_COMPUTE;
		newInsn.output1 = Node::getDataID(tokens[1]);
		newInsn.input1  = Node::getDataID(tokens[2]);
		newInsn.input2  = Node::getDataID(tokens[3]);

		Debug << "SimpleNode: comp " << newInsn.output1 << " " << newInsn.input1 << " " << newInsn.input2 << endl;
	}
	else if (tokens[0].compare("wait") == 0) {
		newInsn.t = SNIT_WAIT;
		Debug << "SimpleNode: wait" << endl;
	}
	else { //Shouldn't happen
		assert(false);
	}

	instructions.push_back(newInsn);
}

//Node is completed when there's no more instructions to do, no more incoming packets,
//and no more outgoing packets to process
bool SimpleNode::isCompleted() {
	return (instructions.size() == 0 
		&& incomingPackets->size() == 0
		&& outgoingPackets.size() == 0);
}
