#include <iostream>
#include <fstream>
using namespace std;

#include "Simulator.h"

//Parses config file
//Sets everything else to NULL
Simulator::Simulator(char *inputConfig) {
	config = new Config(inputConfig);	
	#ifdef DEBUG
		config->writeSettingsToConsole();
	#endif
	nodes = NULL;
	simpleNodes = NULL;
}

Simulator::~Simulator() {
	delete[] simpleNodes;
	//delete[] nodes;
}

//Sets up the topology neighbours and routing table based on config params
bool Simulator::setupTopology() {
	string topology = config->getStr("Topology");

	//Calc numOfNodes
	if (topology.compare("Mesh") == 0 || topology.compare("Torus") == 0) {
		uint radix = config->getUInt("k");
		uint dim = config->getUInt("n");

		numOfNodes = 1;
		for (uint i = 0; i < dim; i++)
			numOfNodes *= radix;		
	}
	else if (topology.compare("Tree") == 0) {
		numOfNodes = config->getUInt("size");
	}
	else {
		cout << "ERROR: Failed to calc numOfNodes. Unsupported topology.\n";
		return false;
	}

	Debug << "Number of Simple Nodes: " << numOfNodes << endl;

	//Init memCntrl node
	memCntrl.setupNode(config, 0);   //nodeId = 0

	//Init topology nodes
	simpleNodes = new SimpleNode[numOfNodes];
	for (uint i = 0; i < numOfNodes; i++) {
		simpleNodes[i].setupNode(config, i+1); //1st nodeId = 1
	}

	// Init the nodes array
	numOfNodes++;
	nodes = new Node*[numOfNodes];
	nodes[0] = &memCntrl;
	for (uint i = 1; i <= numOfNodes; i++) {
		nodes[i] = simpleNodes + (i - 1);
	}

	//Setup the topology neighbours
	if      (topology.compare("Mesh") == 0)  setupMeshNeighbours(1, numOfNodes-1);
	else if (topology.compare("Torus") == 0) setupTorusNeighbours(1, numOfNodes-1);

	else if (topology.compare("Tree") == 0) { 
		setupTreeNeighbours(1, numOfNodes-1);
	}
	else {
		cout << "ERROR: Failed to setup neighbours.\n";
		return false;
	}

	setupMemCntrlNeighbours();

	//Finally, setup the routing table using Dijkstra
	setupRoutingTable();

	return true;
}

void Simulator::setupMemCntrlNeighbours() {
	string str = config->getStr("MemoryCntrl->ConnectToNodes");
	char *cstr = (char *) str.c_str();

	uint start = 1, nodeIndex;
	for (uint curr=1; cstr[curr] != '\0'; curr++) {
		if (cstr[curr] == ',' || cstr[curr] == ']') {
			cstr[curr] = '\0';

			if (strcmp("torRow1", cstr + start) == 0) {
				uint dim = config->getUInt("k");
				for (nodeIndex = 1; nodeIndex <= dim; nodeIndex++) {
					nodes[0]->addNeighbour(nodes[nodeIndex]);
					nodes[nodeIndex]->addNeighbour(nodes[0]);
				}
			}
			else {
				nodeIndex = atoi(cstr + start);
				nodes[0]->addNeighbour(nodes[nodeIndex]);
				nodes[nodeIndex]->addNeighbour(nodes[0]);
				Debug << "Start: " << start << " Curr: " << curr << " NodeIndex: " << nodeIndex << endl;
			}			

			start = curr + 1;
		}
	}
}

void Simulator::setupMeshNeighbours(uint startId, uint numOfNodes) {
	cout << "Mesh Topology to be implemented.\n";
	exit(0);
	//TODO: This is easy as it is simply a subset of the Torus below =]
}

void Simulator::setupTorusNeighbours(uint startId, uint numOfNodes) {
	Debug << "Setting up Torus neighbours.\n";

	uint radix = config->getUInt("k");
	uint dim = config->getUInt("n");

	Debug << "Radix " << radix << " Dim " << dim << endl;
	assert(radix > 0 && dim > 0 && dim <= 2);

	for (uint nodeId=0; nodeId < numOfNodes; nodeId++) {
		Debug << "Node: " << startId + nodeId << endl;

		#ifdef DEBUG
			//north neighbour
			if (nodeId / radix == 0) Debug << "North: " << startId + numOfNodes - radix + nodeId;
			else Debug << "North: " << startId + nodeId - radix;

			//south neighbour
			if (nodeId / radix == radix-1) debug << " South: " << startId + nodeId + radix - numOfNodes;
			else debug << " South: " << startId + nodeId + radix;

			//east neighbour
			if (nodeId % radix == radix-1) debug << " East: " << startId + nodeId - radix + 1;
			else debug << " East: " << startId + nodeId + 1;

			//west neighbour
			if (nodeId % radix == 0) debug << " West: " << startId + nodeId + radix - 1 << endl;
			else debug << " West: " << startId + nodeId - 1 << endl;
		#endif

		//add the routing table entries
		//north neighbour
		if (nodeId / radix == 0) nodes[nodeId + startId]->addNeighbour(nodes[startId + numOfNodes - radix + nodeId]);
		else nodes[nodeId + startId]->addNeighbour(nodes[startId + nodeId - radix]);

		//south neighbour
		if (nodeId / radix == radix-1) nodes[nodeId + startId]->addNeighbour(nodes[startId + nodeId + radix - numOfNodes]);
		else nodes[nodeId + startId]->addNeighbour(nodes[startId + nodeId + radix]);

		//east neighbour
		if (nodeId % radix == radix-1) nodes[nodeId + startId]->addNeighbour(nodes[startId + nodeId - radix + 1]);
		else nodes[nodeId + startId]->addNeighbour(nodes[startId + nodeId + 1]);

		//west neighbour
		if (nodeId % radix == 0) nodes[nodeId + startId]->addNeighbour(nodes[startId + nodeId + radix - 1]);
		else nodes[nodeId + startId]->addNeighbour(nodes[startId + nodeId - 1]);
	}
}

void Simulator::setupTreeNeighbours(uint startId, uint numOfNodes) {
	Debug << "Setting up Tree neighbours.\n";

	for (uint nodeId=0; nodeId < numOfNodes; nodeId++) {
		Debug << "Node: " << nodeId;

		//child 1
		if (nodeId*2 + 1 < numOfNodes) {
			nodes[nodeId + startId]->addNeighbour(nodes[startId + nodeId*2 + 1]);
			debug << " Child1: " << nodeId*2 + 1;
		}

		//child 2
		if (nodeId*2 + 2 < numOfNodes) {
			nodes[nodeId + startId]->addNeighbour(nodes[startId + nodeId*2 + 2]);
			debug << " Child2: " << nodeId*2 + 2;
		}

		//parent
		int parentId = (static_cast<int>(nodeId) - 1) / 2;
		if (nodeId != 0 && parentId >= 0) {
			nodes[nodeId + startId]->addNeighbour(nodes[startId + parentId]);
			debug << " Parent: " << parentId << " ";
		}

		debug << endl;
	}
}

void Simulator::setupRoutingTable() {
	Debug << "Setting up Routing Table\n";

	DijkstraInfo dInfo;
	dInfo.minDistArray = new uint[numOfNodes];
	dInfo.previousNodeIdArray = new uint[numOfNodes];

	for (uint srcId=0; srcId < numOfNodes; srcId++) {
		//run the Dijkstra algorithm on each node
		Dijkstra(nodes[srcId], dInfo);

		//Populate the routing table
		for (uint destId=0; destId < numOfNodes; destId++) {
			if (srcId == destId)
				continue;

			uint nextHop = destId;			
			while (dInfo.previousNodeIdArray[nextHop] != srcId) {
				nextHop = dInfo.previousNodeIdArray[nextHop];
			}

			Debug << "Adding Routing Entry: Dest " << destId << " NextHop " << nextHop << endl;
			if (!nodes[srcId]->isEntryInRoutingTable(destId))
				nodes[srcId]->addRoutingEntry(destId, nodes[nextHop]);
		}
	}

	delete[] dInfo.minDistArray;
	delete[] dInfo.previousNodeIdArray;
}

void Simulator::Dijkstra(Node *src, DijkstraInfo &dInfo) {
	Debug << "In Dijkstra - Source Node: " << src->getNodeId() << endl;
	DijkstraQueueState *nodeQueueState = new DijkstraQueueState[numOfNodes];

	//set all nodes distance from src to MAX value
	for (uint nodeId=0; nodeId < numOfNodes; nodeId++) {
		dInfo.minDistArray[nodeId] = UINT_MAX;
		dInfo.previousNodeIdArray[nodeId] = UINT_MAX;
		nodeQueueState[nodeId] = INQUEUE;
	}

	//set source node's distance to 0
	dInfo.minDistArray[src->getNodeId()] = 0;

	uint queueSize = numOfNodes;
	while (queueSize > 0) {
		uint minNodeId = getNodeIdWithSmallestDist(dInfo.minDistArray, nodeQueueState);
		nodeQueueState[minNodeId] = NOTINQUEUE;
		queueSize--;

		if (dInfo.minDistArray[minNodeId] == UINT_MAX)
			break;

		for (uint nodeId=0; nodeId < numOfNodes; nodeId++) {
			if (nodeQueueState[nodeId] == INQUEUE && nodes[minNodeId]->isNeighbour(nodes[nodeId])) {
				uint altDist = dInfo.minDistArray[minNodeId] + 1; //1 is the distance between neighbouring nodes for now at least

				if (altDist < dInfo.minDistArray[nodeId]) {
					dInfo.minDistArray[nodeId] = altDist;
					dInfo.previousNodeIdArray[nodeId] = minNodeId;
				}
			}
		}
	}

	#ifdef DEBUG
		Debug << "MinDistArray: {";
		for (uint nodeId=0; nodeId < numOfNodes; nodeId++)
			debug << dInfo.minDistArray[nodeId] << ", ";
		debug << "}\n";

		Debug << "PreviousNodeIdArray: {";
		for (uint nodeId=0; nodeId < numOfNodes; nodeId++)
			debug << dInfo.previousNodeIdArray[nodeId] << ", ";
		debug << "}\n";
	#endif

	delete[] nodeQueueState;
}

uint Simulator::getNodeIdWithSmallestDist(uint *minDistArray, DijkstraQueueState *nodeQueueState) {
	uint minNodeId = UINT_MAX;

	for (uint nodeId=0; nodeId < numOfNodes; nodeId++) {
		if (nodeQueueState[nodeId] == INQUEUE && minDistArray[nodeId] < minDistArray[minNodeId])
			minNodeId = nodeId;
	}

	return (minNodeId);
}


void Simulator::readTrafficFile() {
	// Open the file
	ifstream inputFile(config->getStr("TrafficFile").c_str());	
	uint currentNode = -1;

	// Make sure the file is opened
	if (!inputFile.is_open()) {
		printf("ERROR: Can't open file %s\n", config->getStr("TrafficFile").c_str());
		exit(-1);
	}

	string line;
	getline(inputFile, line);
	Debug << line << endl;

	while (inputFile.good()) {
		if (line.compare("") == 0) {
			// Do nothing, it is an empty line
		}
		else if (line.substr(0,2).compare("//") == 0) {
			// Do nothing, it is a comment
		}
		else if (line.substr(0,5).compare("#node") == 0) {
			// It is a directive indicating the following instructions are for which node
			currentNode = atoi(line.substr(6,line.length()).c_str());
		}
		else {
			// it is an instruction
			nodes[currentNode]->addInstruction(line);
		}

		getline(inputFile, line);
		Debug << line << endl;
	}
}

void Simulator::runSimulation() {
	printf("Simulating with these config settings:\n");
	config->writeSettingsToConsole();

	uint cycleCount = 0;
	uint outputFrequency = config->getUInt("OutputFrequency");
	bool simulationComplete;

	do {
		Debug << "Cycle " << cycleCount + 1 << endl;

		// Do the setup before each step
		for (uint nodeId = 0; nodeId < numOfNodes; nodeId++) {
			nodes[nodeId]->preStep();
		}

		// Processing one step for each node
		for (uint nodeId = 0; nodeId < numOfNodes; nodeId++) {
			nodes[nodeId]->step();
		}

		cycleCount++;

		// Check if the simulation is completed
		simulationComplete = true;
		for (uint nodeId=0; simulationComplete && nodeId < numOfNodes; nodeId++) {
			simulationComplete = nodes[nodeId]->isCompleted();
		}

		if (cycleCount % outputFrequency == 0) {
			printf("\tOn Cycle: %d\n",cycleCount);
			// Print out the stat collected so far and reset them
			for (uint nodeId = 0; nodeId < numOfNodes; nodeId++) {
				nodes[nodeId]->printStat();
				nodes[nodeId]->resetCounters();
			}

		}
	}
	while (!simulationComplete);

	printf("\nSimulation Results\n");
	printf("\tCycle Count: %d\n", cycleCount);
}
