#include <iostream>
#include <sstream>
using namespace std;

#include "SerialSolver.h"
#include "../Parser/Parser.h"
#include "../Validator/Validator.h"
#include "../GeneratePeers/GeneratePeers.h"

/* STATIC MEMBERS */
map<uint, uint> SudokuSerialSolver::ValueLookup;

uint SudokuSerialSolver::numPeers = 0;
uint **SudokuSerialSolver::peers = NULL;

/**
 * Mapping from possibilities when numPossible = 1 -> actual cell value
 * So For 9x9 puzzle, 0b000000100 -> 3
 * he LookupTable becomes as large as it needs to be.
 * Can be called multiple times simply to extend the table.
 */
void SudokuSerialSolver::populateLookupTable(uint size) {
	for (uint i = ValueLookup.size(); i < size; i++)
		ValueLookup[1 << i] = i + 1;
}

map<SudokuSerialSolver::Cell*, uint> SudokuSerialSolver::cellMap;

string SudokuSerialSolver::cellOut(uint row, uint col) {
	ostringstream convert;
	convert << "cell[" << row << "][" << col << "]v";
	return convert.str();
}

/* INSTANCE MEMBERS */

//Initialize an unsolved Sudoku Grid and other data structures needed to solve it.
SudokuSerialSolver::SudokuSerialSolver(char *inPuzzleFile, char *outInsnFilePath) {
	if (outInsnFilePath == NULL) {
		outInsnFile = NULL;
	}
	else {
		outInsnFile = new ofstream(outInsnFilePath);
		if (!outInsnFile->is_open()) {
			cout << "Unable to open output insn file for writing.\n";
			exit(1);
		}		
	}

	//Read the puzzle from a file; size is updated
	//memory is allocated for puzzle inside the function called
	uint *puzzle = ReadSudokuFromFile(inPuzzleFile, size);

	if (outInsnFile)
		*outInsnFile << "//Sudoku Parameters Used:"
		             << "\n//Size = " << size << endl;

	//extend Lookup table with new values as needed
	populateLookupTable(size);

	//generate peers
	if (peers == NULL)
		peers = GeneratePeers(size, numPeers);
	
	numSingletons = 0;
	cells = new Cell[size*size];
	
	for (uint cellIndex=0; cellIndex<size*size; cellIndex++) {
		//value of 0 means that cell is unsolved
		if (puzzle[cellIndex] == 0) {
			cells[cellIndex].numPossible = size;
			cells[cellIndex].possibilities = (1 << size) - 1;
		}
		//the cell has been solved already
		else {
			cells[cellIndex].numPossible = 1;
			cells[cellIndex].possibilities = 1 << (puzzle[cellIndex] - 1);
			numSingletons++;
		}
	}

	delete[] puzzle;
}

SudokuSerialSolver::~SudokuSerialSolver() {
	delete[] cells;

	if (outInsnFile) {
		outInsnFile->close();
		delete outInsnFile;
	}
}

uint *SudokuSerialSolver::createPuzzleGrid() {
	uint *puzzle = new uint[size*size];

	for (uint cellIndex=0; cellIndex<size*size; cellIndex++) {
		if (cells[cellIndex].numPossible != 1)
			puzzle[cellIndex] = 0;
		else
			puzzle[cellIndex] = ValueLookup[cells[cellIndex].possibilities];
	}

	return (puzzle);
}

//Print puzzle to screen
void SudokuSerialSolver::printPuzzleToScreen() {
	uint *puzzle = createPuzzleGrid();	
	WriteSudokuToConsole(puzzle, size);

	delete[] puzzle;
}

void SudokuSerialSolver::printPuzzleToFile(char* fileName) {
	uint *puzzle = createPuzzleGrid();
	WriteSudokuToFile(fileName, puzzle, size);

	delete[] puzzle;
}

void SudokuSerialSolver::doConstraintPropagation() {
	bool iterChangedAnything;

	do {
		Debug << "Another Iteration\n";
		iterChangedAnything = false;

		//For each non-singleton cell:
		for (uint cellIndex=0; cellIndex<size*size; cellIndex++) {
			if (cells[cellIndex].numPossible == 1)
				continue;

			//For each of its peers:
			//If peer is singleton, remove it from my possibility list
			for (uint peer=0; peer<numPeers; peer++) {
				uint peerIndex = peers[cellIndex][peer];

				if (cells[peerIndex].numPossible == 1) {
					if (outInsnFile) {
						//read operation on peer
						*outInsnFile << "comp " << cellOut(peerIndex/size, peerIndex%size) << cellMap[cells + peerIndex]
						             << " "     << cellOut(peerIndex/size, peerIndex%size) << cellMap[cells + peerIndex]
							         << " "     << "nil"                                   << cellMap[NULL] << endl;
					}

					//If it's the first time we're eliminating this possibility,
					//decrement num_possible for this cell
					if ((cells[cellIndex].possibilities & cells[peerIndex].possibilities) != 0) {
						cells[cellIndex].numPossible--;
						cells[cellIndex].possibilities = cells[cellIndex].possibilities & ~cells[peerIndex].possibilities;
						iterChangedAnything = true;

						//write operation on curr cell
						if (outInsnFile) {
							*outInsnFile << "comp " << cellOut(cellIndex/size, cellIndex%size) << cellMap[cells + cellIndex]+1
								         << " "     << cellOut(cellIndex/size, cellIndex%size) << cellMap[cells + cellIndex]
							             << " "     << cellOut(peerIndex/size, peerIndex%size) << cellMap[cells + peerIndex] << endl;

							cellMap[cells + cellIndex]++;
						}

						Debug << "Changed - cell " << cellIndex << " peer: " << peerIndex
						      << " Removed: " << ValueLookup[cells[peerIndex].possibilities] << "\n";
					}
				}
			}

			debug << "\n";

			// if the cell was successfully solved
			if (cells[cellIndex].numPossible == 1)
				numSingletons++;
		}
	} while (iterChangedAnything);
}

void SudokuSerialSolver::solve() {
	Debug << "Num of singletons before: " << numSingletons << endl;
	doConstraintPropagation();

	if (outInsnFile) {
		uint cellIndex;
		map<Cell *, uint>::const_iterator elem;

		for (elem = cellMap.begin(); elem != cellMap.end(); elem++) {
			if (elem->second > 0) {
				cellIndex = elem->first - cells;
				*outInsnFile << "send 0 " << cellOut(cellIndex/size, cellIndex%size) << elem->second << endl;
			}
		}
	}
}

bool SudokuSerialSolver::verify(uint& errRow, uint& errCol) {
	uint *puzzle = createPuzzleGrid();
	bool retVal = ValidateSolvedSudoku(puzzle, size, errRow, errCol);

	delete[] puzzle;
	return retVal;
}

bool SudokuSerialSolver::verifyUnsolved(uint& errRow, uint& errCol) {
	uint *puzzle = createPuzzleGrid();
	bool retVal = ValidateUnsolvedSudoku(puzzle, size, errRow, errCol);

	delete[] puzzle;
	return retVal;
}
